import os

BANYAN_PYTHON_BRANCH_NAME = "v22.05.14"
BANYAN_PYTHON_PACKAGES = ["banyan"]
BANYAN_API_ENDPOINT = os.getenv(
    "BANYAN_API_ENDPOINT",
    default="https://4whje7txc2.execute-api.us-west-2.amazonaws.com/prod/",
)



from tqdm import tqdm
import urllib

from .config import configure
from .imports import *
from .session import get_cluster_name
from .utils import (
    send_request_get_response,
    get_aws_config_region,
    s3_bucket_arn_to_name,
)


s3 = boto3.client("s3")
clusters = dict()


def create_cluster(
    name: str = None,
    instance_type: str = "m4.4xlarge",
    max_num_workers: str = 2048,
    initial_num_workers: int = 16,
    min_num_workers: int = 0,
    iam_policy_arn: str = None,
    s3_bucket_arn: str = None,
    s3_bucket_name: str = None,
    scaledown_time: int = 25,
    region: str = None,
    vpc_id: str = None,
    subnet_id: str = None,
    nowait: bool = False,
    ec2_key_pair: str = None,
    **kwargs,
):
    """Creates a new cluster or re-creates a previously destroyed cluster.
    If no vpc_id and subnet_id are provided,
    the cluster is by default created in the default public VPC
    and subnet in your AWS account.
    """
    global clusters

    # Configure using parameters
    c = configure(**kwargs)

    clusters_local = get_clusters(**kwargs)
    if name is None:
        name = "Cluster " + str(len(clusters_local) + 1)
    if region is None:
        region = get_aws_config_region()

    # Check if the configuration for this cluster name already exists
    # If it does, then recreate cluster
    if name in clusters_local:
        if clusters_local[name].status == "terminated":
            logging.info(f"Started re-creating cluster named {name}")
            send_request_get_response(
                "create-cluster", {"cluster_name": name, "recreate": True}
            )
            if not nowait:
                wait_for_cluster(name)
            # Cache info
            return get_cluster(name)
        else:
            raise Exception(
                f"Cluster with name {name} already exists and its current status is {str(clusters_local[name].status)}"
            )

    # Construct arguments
    if s3_bucket_name is not None:
        s3_bucket_arn = f"arn:aws:s3:::{s3_bucket_name}"
    elif s3_bucket_arn is not None:
        s3_bucket_name = s3_bucket_arn.split(":")[-1]

    if s3_bucket_arn is None:
        s3_bucket_arn = ""
    elif s3_bucket_name not in s3.list_buckets():
        logging.error(
            f"Bucket {s3_bucket_name} does not exist in connected AWS account"
        )

    # Construct cluster creation
    cluster_config = {
        "cluster_name": name,
        "instance_type": instance_type,
        "max_num_workers": max_num_workers,
        "initial_num_workers": initial_num_workers,
        "min_num_workers": min_num_workers,
        "aws_region": region,
        "s3_read_write_resource": s3_bucket_arn,
        "scaledown_time": scaledown_time,
        "recreate": False,
    }

    if "ec2_key_pair_name" in c["aws"]:
        cluster_config["ec2_key_pair"] = c["aws"]["ec2_key_pair_name"]
    if iam_policy_arn is not None:
        cluster_config["additional_policy"] = iam_policy_arn
    if vpc_id is not None:
        cluster_config["vpc_id"] = vpc_id
    if subnet_id is not None:
        cluster_config["subnet_id"] = subnet_id
    if not ec2_key_pair == None:
        cluster_config["ec2_key_pair"] = ec2_key_pair

    logging.info(f"Started creating cluster named {name}")
    # Send request to create cluster
    send_request_get_response("create-cluster", cluster_config)

    if not nowait:
        wait_for_cluster(name)

    # Cache info
    get_cluster(name)

    return clusters[name]


def destroy_cluster(name: str, **kwargs):
    configure(**kwargs)
    logging.info(f"Destroying cluster named {name}")
    send_request_get_response("destroy-cluster", {"cluster_name": name})


def delete_cluster(name: str, **kwargs):
    configure(**kwargs)
    logging.info(f"Deleting cluster named {name}")
    send_request_get_response(
        "destroy-scluster", {"cluster_name": name, "permanently_delete": True}
    )


def update_cluster(name: str, **kwargs):
    configure(**kwargs)
    logging.info(f"Updating cluster named {name}")
    send_request_get_response("update-cluster", {"cluster_name": name})


def assert_cluster_is_ready(name: str, **kwargs):
    logging.info(f"Setting status of cluster named {name} to running")
    configure(**kwargs)
    send_request_get_response("set-cluster-ready", {"cluster_name": name})


class Cluster:
    def __init__(
        self, name: str, status: str, status_explanation: str, s3_bucket_arn: str
    ):
        self.name = name
        self.status = status
        self.status_explanation = status_explanation
        self.s3_bucket_arn = s3_bucket_arn


def get_clusters(cluster_name=None, **kwargs):
    logging.debug("Downloading description of clusters")
    filters = {}
    if cluster_name is not None:
        filters["cluster_name"] = cluster_name
    response = send_request_get_response("describe-clusters", {"filters": filters})
    clusters_dict = {
        name: Cluster(
            name,
            c["status"],
            c["status_explanation"] if "status_explanation" in c else "",
            c["s3_read_write_resource"],
        )
        for (name, c) in response["clusters"].items()
    }

    # Cache info
    global clusters
    for (name, c) in clusters_dict.items():
        clusters[name] = c

    return clusters_dict


def get_cluster_s3_bucket_arn(cluster_name=None, **kwargs):
    configure(**kwargs)
    if cluster_name is None:
        cluster_name = get_cluster_name()
    global clusters
    # Check if cached, since this property is immutable
    if cluster_name not in clusters:
        get_cluster(cluster_name)
    return clusters[cluster_name].s3_bucket_arn


def get_cluster_s3_bucket_name(cluster_name=None, **kwargs):
    configure(**kwargs)
    if cluster_name is None:
        cluster_name = get_cluster_name()
    return s3_bucket_arn_to_name(get_cluster_s3_bucket_arn(cluster_name))


def get_cluster(name: str = None, **kwargs):
    if name is None:
        name = get_cluster_name()
    return get_clusters(name, **kwargs)[name]


def get_running_clusters(*args, **kwargs):
    return dict(
        filter(
            lambda entry: entry[1].status == "running",
            get_clusters(*args, **kwargs).items(),
        )
    )


def get_cluster_status(name: str = None, **kwargs):
    if name is None:
        name = get_cluster_name()
    global clusters
    if name in clusters:
        if clusters[name].status == "failed":
            logging.error(clusters[name].status_explanation)
    # If it is not failed, then retrieve status, in case it has changed
    c = get_cluster(name, **kwargs)
    if c.status == "failed":
        raise Exception(c.status_explanation)
    return c.status


def wait_for_cluster(name: str = None, **kwargs):
    if name is None:
        name = get_cluster_name()
    t = 5
    i = 0
    cluster_status = get_cluster_status(name, **kwargs)
    while cluster_status == "creating" or cluster_status == "updating":
        if cluster_status == "creating":
            pbar = tqdm(desc=f"Setting up cluster {name}")
        else:
            pbar = tqdm(desc=f"Updating cluster {name}")
        time.sleep(t)
        if t < 80:
            t *= 2
        cluster_status = get_cluster_status(name, **kwargs)
        pbar.update(i)
        i += 1
    try:
        pbar.close()
    except:
        pass
    if cluster_status == "running":
        logging.info(f"Cluster {name} is ready")
    elif cluster_status == "terminated":
        raise Exception(f"Cluster {name} no longer exists")
    elif cluster_status not in ["creating", "updating"]:
        raise Exception(f"Failed to set up cluster named {name}")
    else:
        raise Exception(f"Cluster {name} has unexpected status: {cluster_status}")


def upload_to_s3(src_path, dst_name=None, cluster_name=None, **kwargs):
    if dst_name is None:
        dst_name = os.path.basename(src_path)
    if cluster_name is None:
        cluster_name = get_cluster_name()

    configure(**kwargs)
    bucket_name = get_cluster_s3_bucket_name(cluster_name)

    if src_path.startswith("http://") or src_path.startswith("https://"):
        with urllib.request.urlopen(src_path) as f:
            s3.upload_fileobj(f, bucket_name, dst_name)
    if src_path.startswith("s3://"):
        bucket_src, key_src = src_path.replace("s3://", "").split("/", 1)
        copy_source = {"Bucket": bucket_src, "Key": key_src}
        s3.meta_client.copy(copy_source, bucket_name, dst_name)
    else:
        if src_path.startswith("file://"):
            src_path = src_path[8:]
        src_path = Path(src_path)

        if src_path.is_file():
            s3.meta.client.upload_file(src_path, bucket_name, dst_name)
        else:
            files = [
                f
                for f in os.listdir(src_path)
                if os.path.isfile(os.path.join(src_path, f))
            ]
            for file in files:
                s3.meta.client.upload_file(
                    Path(os.path.join(src_path, file)),
                    bucket_name,
                    os.path.join(dst_name, file),
                )
    return dst_name



import boto3
import json

from .id import generate_message_id, ResourceId
from .sessions import get_session, get_session_id, end_session
from .utils import from_py_value_contents, to_py_value_contents

sqs = boto3.client("sqs")

#################
# GET QUEUE URL #
#################


def get_scatter_queue(resource_id=None):
    if resource_id is None:
        resource_id = get_session().resource_id
    return sqs.get_queue_url(QueueName="banyan_" + resource_id + "_scatter.fifo")[
        "QueueUrl"
    ]


def get_gather_queue(resource_id=None):
    if resource_id is None:
        resource_id = get_session().resource_id
    return sqs.get_queue_url(QueueName="banyan_" + resource_id + "_gather.fifo")[
        "QueueUrl"
    ]


def get_execution_queue(resource_id=None):
    if resource_id is None:
        resource_id = get_session().resource_id
    return sqs.get_queue_url(QueueName="banyan_" + resource_id + "_execution.fifo")[
        "QueueUrl"
    ]


###################
# RECEIVE MESSAGE #
###################


def get_next_message(queue, delete=True):
    message_receiving_result = sqs.receive_message(
        QueueUrl=queue, MaxNumberOfMessages=1
    )
    m = (
        message_receiving_result["Messages"][0]["ReceiptHandle"]
        if len(message_receiving_result["Messages"]) > 0
        else None
    )

    while m is None:
        message_receiving_result = sqs.receive_message(
            QueueUrl=queue, MaxNumberOfMessages=1
        )
        m = (
            message_receiving_result["Messages"][0]["ReceiptHandle"]
            if len(message_receiving_result["Messages"]) > 0
            else None
        )

    if delete:
        sqs.delete_message(QueueUrl=queue, ReceiptHandle=m)
    return message_receiving_result["Messages"][0]["Body"]


def receive_next_message(queue_name):
    content = get_next_message(queue_name)

    if content.startswith("JOB_READY") or content.startswith("SESSION_READY"):
        response = {"kind": "SESSION_READY"}
        return response
    elif content.startswith("EVALUATION_END"):
        response = {"kind": "EVALUATION_END"}
        response["end"] = content.endswith("MESSAGE_END")
        print(content[14:-1])
        return response
    elif content.startswith("JOB_FAILURE") or content.startswith(
        content, "SESSION_FAILURE"
    ):
        tail = 11 if content.endswith("MESSAGE_END") else 0
        head_len = 11 if content.startswith("JOB_FAILURE") else 15
        # This print statement is needed, so that we can print out the error message
        print(content[head_len:tail])
        if content.endswith("MESSAGE_END"):
            end_session(
                failed=True, release_resources_now=content.startswith("JOB_FAILURE")
            )
            raise RuntimeError("Session failed; see preceding output")
        response = {"kind": "SESSION_FAILURE"}
        return response
    else:
        return json.loads(content)


# Used by Banyan/src/pfs.jl, intended to be called from the executor
def receive_from_client(value_id):
    # Send scatter message to client
    send_message(
        get_gather_queue(),
        json.dumps({"kind": "SCATTER_REQUEST", "value_id": value_id}),
    )
    # Receive response from client
    m = json.loads(get_next_message(get_scatter_queue()))
    v = from_py_value_contents(m["contents"])
    return v


################
# SEND MESSAGE #
################


def send_message(queue_name, message):
    # queue_url = sqs.get_queue_url(queue_name)
    return sqs.send_message(
        QueueUrl=queue_name,  # queue_url,
        MessageBody=message,  # TODO: Is that correct?,
        MessageGroupId="1",
        MessageDeduplicationId=generate_message_id(),  # TODO: where does that function come from?
    )


def send_to_client(value_id, value):
    print("QUEUE NAME: ", get_gather_queue())
    send_message(
        get_gather_queue(),
        json.dumps(
            {
                "kind": "GATHER",
                "value_id": value_id,
                "contents": to_py_value_contents(value),
            }
        ),
    )


###########################
# GET MESSAGES FROM QUEUE #
###########################


##########################
# SEND MESSAGES TO QUEUE #
##########################



from .constants import BANYAN_API_ENDPOINT
from .imports import *
import boto3
import codecs
import hashlib
import inspect
import json
import logging
import os
import pickle
import platform
import pytz
import requests
import toml


from botocore.exceptions import ClientError
from .config import load_config
from datetime import datetime
from typing import Dict

s3_client = boto3.client("s3")

s3 = boto3.client("s3")


def get_aws_config_region():
    return s3.meta.region_name


def method_to_endpoint(method):
    return method.replace("_", "-")


def s3_bucket_arn_to_name(s3_bucket_arn: str):
    # Get s3 bucket name from arn
    s3_bucket_name = s3_bucket_arn.split(":")[-1]
    if s3_bucket_name.endswith("/") or s3_bucket_name.endswith("*"):
        s3_bucket_name = s3_bucket_name[:-1]
    elif s3_bucket_name.endswith("/*"):
        s3_bucket_name = s3_bucket_name[:-2]
    return s3_bucket_name


def send_request_get_response(method: str, content: dict):
    configuration = load_config()
    user_id = configuration["banyan"]["user_id"]
    api_key = configuration["banyan"]["api_key"]

    url = (BANYAN_API_ENDPOINT) + method_to_endpoint(method)
    content["debug"] = True
    headers = {
        "content-type": "application/json",
        "Username-APIKey": f"{user_id}-{api_key}",
    }
    resp = requests.post(url=url, json=content, headers=headers)
    data = json.loads(resp.text)
    if resp.status_code == 403:
        raise Exception(
            "Please use a valid user ID and API key. Sign into the dashboard to retrieve these credentials."
        )
    elif resp.status_code == 504:
        # HTTP request timed out, for example
        if isinstance(data, Dict) and "message" in data:
            data = data["message"]
        # @error data #?????
        return None
    elif resp.status_code == 500 or resp.status_code == 504:
        raise Exception(data)
    elif resp.status_code == 502:
        raise Exception("Sorry there has been an error. Please contact support.")
    return data


def is_debug_on():
    return logging.DEBUG >= logging.root.level


def get_python_version():
    """Gets the python version.

    Returns
    -------
    string
        Python version is returned as a string
    """
    return platform.python_version()


def parse_time(time):
    """Converts given time to local timezone.

    Parameters
    ---------
    time : string
        The current time in the format "yyyy-mm-dd-HH:MM:SSzzzz"
    Returns
    -------
    string
        The DateTime is returned.
    """
    time = datetime.fromisoformat(time)
    # time = datetime.strptime(time[:-4], '%Y-%m-%d-%H:%M:%S') #we don't want milli-second
    timezone = pytz.timezone("UTC")
    time = timezone.localize(time)
    local_time = time.astimezone()
    return local_time


# TO DO - to test this function
def get_loaded_packages():
    """Returns all the packages/libraries that are currently imported by the user"""
    return [
        p[0]
        for p in locals().items()
        if inspect.ismodule(p[1]) and not p[0].startswith("__")
    ]


def get_hash(s):
    """Gets a unique represetation of a string

    Parameters
    ----------
    s : string

    Returns
    -------
    string :
        the SHA256 hash of the string

    """
    hs = hashlib.sha256(s.encode("utf-8")).hexdigest()
    return hs


def upload_file_to_s3(filename, bucket, object_name=None):
    """Uploads file to the S3 bucket

    Parameters
    ----------
    filename: string
        Is the local path to the file to upload
    bucket: string
        Is the S3 bucket to which to upload the file to

    Returns
    -------
    boolean: True if file was uploaded, else False
    """

    # if S3 object_name not specified, use filename
    key = os.path.basename(filename)

    # upload the file
    try:
        reponse = s3_client.upload_file(
            filename, bucket, object_name if object_name is not None else key
        )
    except ClientError as e:
        logging.error(e)
        return False
    return True


def load_toml(path):
    # path --> "file://.banyan/banyanconfig.toml"
    # path[2:], path[4:8], path[:9]
    if isinstance(path, list):
        result = {}
        for p in path:
            result.update(load_toml(p))
        return result
    if path.startswith("file://"):
        return toml.load(path[7:-1])

    elif path.startswith("s3://"):
        raise Exception("S3 path not currently supported")

    elif (path.startswith("http://")) or (path.startswith("https://")):
        r = (
            requests.get(path)
        ).content  # downloads the data from the internet into a toml-fomatted string
        return toml.loads(requests.get(path).text)


def to_py_value_contents(py):
    # Handle functions defined in a module
    # TODO: Document this special case
    # if jl isa Function && !(isdefined(Base, jl) || isdefined(Core, jl) || isdefined(Main, jl))
    # if jl isa Expr && eval(jl) isa Function
    #     jl = Dict("is_banyan_udf" => true, "code" => jl)
    # end

    # Convert Python object to string
    return codecs.encode(pickle.dumps(py), "base64").decode()


def from_py_value_contents(py_value_contents):
    # Converty string to Python object
    return pickle.loads(codecs.decode(py_value_contents.encode(), "base64"))

    # # Handle functions defined in a module
    # if res isa Dict && haskey(res, "is_banyan_udf") && res["is_banyan_udf"]
    #     eval(res["code"])
    # else
    #     res
    # end



__version__ = "0.1.0"
from .constants import *
from .clusters import (
    create_cluster,
    destroy_cluster,
    delete_cluster,
    update_cluster,
    assert_cluster_is_ready,
    Cluster,
    get_clusters,
    get_cluster_s3_bucket_arn,
    get_cluster_s3_bucket_name,
    get_cluster,
    get_running_clusters,
    get_cluster_status,
    wait_for_cluster,
    upload_to_s3,
)
from .config import load_config, write_config, configure
from .id import generate_message_id
from .queues import (
    get_scatter_queue,
    get_gather_queue,
    get_execution_queue,
    get_next_message,
    receive_next_message,
    receive_from_client,
    send_message,
    send_to_client,
)
from .session import Session, set_session, get_session_id, get_session, get_cluster_name
from .sessions import (
    end_session,
    end_all_sessions,
    get_sessions,
    get_running_sessions,
    get_session_status,
    download_session_logs,
    wait_for_session,
    start_session,
    run_session,
)
from .utils import (
    get_aws_config_region,
    send_request_get_response,
    is_debug_on,
    get_python_version,
    parse_time,
    get_loaded_packages,
    get_hash,
    upload_file_to_s3,
    load_toml,
)



ResourceId = str

generated_message_ids = set()
num_message_ids_issued = 0

def generate_message_id():
    global generated_message_ids
    global num_message_ids_issued
    num_message_ids_issued += 1
    v = str(num_message_ids_issued)
    generated_message_ids.add(v)
    return v



# Standard Library
import logging
import os
from pathlib import Path
import time
from typing import Any, Dict, Optional, Union

# from _typeshed import NoneType

# Third Party
import boto3
from botocore.config import Config
import requests
import toml

# Constructed
NoneType = type(None)



"""
This file contains utils files related to setting and
loading configurations.
"""

from banyan.imports import *
from copy import deepcopy

banyan_config = dict()  # Global variable representing configuration


def load_config(banyanconfig_path: Optional[str] = None):
    """Loads configuration from given file

    Arguments:
        banyanconfig_path:Optional[str], defaults to None
            filepath to Banyan configuration file
            if None (recommended), will use "$HOME/.banyan/banyanconfig.toml"
    """
    global banyan_config

    if banyan_config is None:
        if banyanconfig_path is None:
            banyanconfig_path = os.path.join(
                os.path.expanduser("~"), ".banyan/banyanconfig.toml"
            )
        if os.path.exists(banyanconfig_path):
            banyan_config = toml.load(banyanconfig_path)
    return banyan_config


def write_config(banyanconfig_path: Optional[str] = None):
    """Writes configuration to given file

    Arguments:
        banyanconfig_path:Optional[str], defaults to None
            filepath to Banyan configuration file
            if None (recommended), will use "$HOME/.banyan/banyanconfig.toml"
    """
    global banyan_config
    if banyanconfig_path is None:
        banyanconfig_path = os.path.join(
            os.path.expanduser("~"), ".banyan/banyanconfig.toml"
        )

    os.makedirs(os.path.join(os.path.expanduser("~"), ".banyan/"), exist_ok=True)

    with open(banyanconfig_path, "w") as f:
        toml.dump(banyan_config, f)
    return banyan_config


def configure(
    user_id: Optional[str] = None,
    api_key: Optional[str] = None,
    ec2_key_pair_name: Optional[str] = None,
    banyanconfig_path: Optional[str] = None,
):
    """Sets configuration.
    If any provided configuration parameter is None,
    will check environment variable, then last used/loaded configuration

    Arguments:
    - user_id:Optional[str], defaults to None
    - api_key:Optional[str], defaults to None
    - ec2_key_pair_name:Optional[str], defaults to None
    - banyanconfig_path:Optional[str], defaults to None
        file to save configurations to
        if None (recommended), will use "$HOME/.banyan/banyanconfig.toml"
    """

    c = load_config(banyanconfig_path)
    global banyan_config
    banyan_config = c

    # Check environment variables
    user_id_env = os.getenv("BANYAN_USER_ID", default=None)
    if user_id is None and not user_id_env is None:
        user_id = user_id_env

    api_key_env = os.getenv("BANYAN_API_KEY", default=None)
    if api_key is None and not api_key_env is None:
        api_key = api_key_env

    ec2_env = os.getenv("BANYAN_EC2_KEY_PAIR_NAME")
    if ec2_key_pair_name is None and not ec2_env == None:
        ec2_key_pair_name = ec2_env

    # Check banyan_config file
    banyan_config_has_info = not (banyan_config is None or banyan_config == {})
    if (
        user_id is None
        and banyan_config_has_info
        and "banyan" in banyan_config
        and "user_id" in banyan_config["banyan"]
    ):
        user_id = banyan_config["banyan"]["user_id"]
    if (
        api_key is None
        and banyan_config_has_info
        and "banyan" in banyan_config
        and "api_key" in banyan_config["banyan"]
    ):
        api_key = banyan_config["banyan"]["api_key"]
    if (
        ec2_key_pair_name is None
        and banyan_config_has_info
        and "aws" in banyan_config
        and "ec2_key_pair_name" in banyan_config["aws"]
    ):
        ec2_key_pair_name = banyan_config["aws"]["ec2_key_pair_name"]

    # Ensure a configuration has been created or can be created. Otherwise,
    # return nothing
    existing_banyan_config = deepcopy(banyan_config)
    if banyan_config is None or banyan_config == {}:
        if not user_id is None and not api_key is None:
            banyan_config = {
                "banyan": {"user_id": user_id, "api_key": api_key},
                "aws": {},
            }
            is_modified = True
        else:
            raise Exception(
                "Your user ID and API key must be specified using either keyword arguments, environment variables, or banyanconfig.toml"
            )
    if not existing_banyan_config == banyan_config:
        write_config(banyanconfig_path)

    return banyan_config



sessions = dict()
current_session_id = None


class Session:
    """Stores information about one session"""

    def __init__(self, cluster_name, session_id, resource_id, nworkers, sample_rate):
        self._cluster_name = cluster_name
        self._session_id = session_id
        self._resource_id = resource_id
        self._nworkers = nworkers
        self._sample_rate = sample_rate
        self._locations = {}
        self._pending_requests = []
        self._futures_on_client = {}

    @property
    def cluster_name(self):
        return self._cluster_name

    @property
    def resource_id(self):
        return self._resource_id

    @property
    def nworkers(self):
        return self._nworkers

    @property
    def sample_rate(self):
        return self._sample_rate

    # Add other getters if needed


def set_session(session_id: str, session=None, *args, **kwargs):
    """Sets the session ID.

    Parameters
    ----------
    session_id : string
        Session ID to use
    session : Session
        If not None (default), the global sessions table is updated to include
        this session.
    """

    global current_session_id
    current_session_id = session_id

    global sessions
    if session is not None:
        sessions[current_session_id] = session


def get_session_id(*args, **kwargs):
    """Returns the value of the global variable set to the current session ID.

    Returns
    -------
    string
        Current session ID
    """

    global current_session_id
    if current_session_id is None:
        raise Exception(
            "No session started or selected using `start_session` or `with_session` or `set_session`. The current session may have been destroyed or no session started yet.",
        )
    return current_session_id


def get_session(session_id=None, *args, **kwargs):
    """Get information about the current session.

    Parameter
    --------
    session_id : string
        Session ID to get information for

    Returns
    -------
    Session
        Information about the given session ID

    Raises
    ------
        Exception if the session ID is for a session that wasn't created by this
        process or has failed
    """

    if session_id is None:
        session_id = get_session_id()
    global sessions  # an empty dictionary that will get filled up with mappings from session_id ->instances of the class Session
    if session_id not in sessions:
        raise Exception(
            f"The selected job with ID {session_id} does not have any information; if it was created by this process, it has either failed or been destroyed."
        )
    return sessions[session_id]


def get_cluster_name(*args, **kwargs):
    """Gets the name of the cluster that the current session is running on.

    Returns
    -------
    string
        Name of the cluster that the current session is running on.
    """

    return get_session().cluster_name



import boto3
import logging
import os
from pygit2 import Repository
import time
from tqdm import tqdm


from .constants import BANYAN_PYTHON_BRANCH_NAME, BANYAN_PYTHON_PACKAGES
from .clusters import get_running_clusters, get_cluster_s3_bucket_name, wait_for_cluster
from .config import configure
from .session import (
    get_session_id,
    set_session,
    sessions,
    current_session_id,
    get_session,
    Session,
)
from .utils import (
    get_hash,
    get_loaded_packages,
    get_python_version,
    load_toml,
    parse_time,
    send_request_get_response,
    upload_file_to_s3,
)


def start_session(
    cluster_name: str = None,
    nworkers: int = 16,
    release_resources_after: int = 20,
    print_logs: bool = False,
    store_logs_in_s3: bool = True,
    store_logs_on_cluster: bool = False,
    log_initialization: bool = False,
    sample_rate: int = None,
    session_name: str = None,
    files: list = None,
    code_files: list = None,
    force_update_files: bool = False,
    pf_dispatch_table: str = None,
    using_modules: list = None,
    # pip_requirements_file = None, # paths to a requirements.txt file that contains packages to be installed with pip
    # conda_environment_file = None, # paths to environment.yml file that contains packages to be installed with conda
    project_dir: str = None,  # a pyproject.toml file containing information about a poetry environment
    url: str = None,
    branch: str = None,
    directory: str = None,
    dev_paths: list = None,
    force_sync: bool = False,
    force_pull: bool = False,
    force_install=False,
    estimate_available_memory=True,
    nowait=False,
    email_when_ready=None,
    for_running=False,
    *args,
    **kwargs,
):
    """
    Starts a new session.
    """

    configure(*args, **kwargs)

    if sample_rate is None:
        sample_rate = nworkers

    if files is None:
        files = []

    if code_files is None:
        code_files = []

    if using_modules is None:
        using_modules = []

    if dev_paths is None:
        dev_paths = []

    if project_dir is None:
        project_dir = os.getcwd()  # gets the current working directory

    poetry_pyproject_file = os.path.join(project_dir, "pyproject.toml")
    poetry_lock_file = os.path.join(project_dir, "poetry.lock")

    # Construct parameters for starting session
    if cluster_name is None:
        # running_clusters is dictionary
        running_clusters = get_running_clusters()
        if len(running_clusters) == 0:
            raise Exception(
                "Failed to start session: you don't have any clusters created"
            )
        else:
            cluster_name = list(running_clusters.keys())[0]
    version = get_python_version()

    session_configuration = {
        "cluster_name": cluster_name,
        "num_workers": nworkers,
        "release_resources_after": release_resources_after,
        "return_logs": print_logs,
        "store_logs_in_s3": store_logs_in_s3,
        "store_logs_on_cluster": store_logs_on_cluster,
        "version": version,
        "benchmark": os.environ.get("BANYAN_BENCHMARK", "0") == "1",
        "main_modules": get_loaded_packages(),
        "using_modules": using_modules,
        "reuse_resources": not force_update_files,
        "estimate_available_memory": estimate_available_memory,
        "language": "py",
    }

    if session_name is None:
        session_configuration["session_name"] = session_name

    if email_when_ready is None:
        session_configuration["email_when_ready"] = email_when_ready

    s3_bucket_name = get_cluster_s3_bucket_name(cluster_name)
    environment_info = {}
    # If a url is not provided, then use the local environment
    if url is None:
        # There are two files we need: pyproject.toml and project.lock
        # Check if the pyproject.toml exists
        if os.path.exists(poetry_pyproject_file):
            # Check if the project.lock exists
            if os.path.exists(poetry_lock_file):
                # Read in the poetry.lock
                with open(poetry_lock_file) as f:
                    poetry_lock_file_contents = f.read()
            else:
                # If it doesn't exist - that's fine..
                poetry_lock_file_contents = ""
            # Read in the pyproject.toml
            with open(poetry_pyproject_file) as f:
                file_contents = f.read()

            # At this point, both files have been read in so we go ahead and
            # get the hash of them concatenated
            environment_hash = get_hash(poetry_lock_file_contents + file_contents)
            environment_info["environment_hash"] = environment_hash

            # Upload the pyproject.toml file to S3
            object_name = environment_hash + "/pyproject.toml"
            upload_file_to_s3(poetry_pyproject_file, s3_bucket_name, object_name)
            environment_info["pyproject_toml"] = object_name

            if poetry_lock_file_contents != "":
                object_name = environment_hash + "/poetry.lock"
                upload_file_to_s3(poetry_lock_file, s3_bucket_name, object_name)
                environment_info["poetry_lock"] = object_name

        else:
            # It has to exist!
            raise Exception("poetry_pyproject_file does not exist")

    else:
        # Otherwise, use url and optionally a particular branch
        environment_info["url"] = url

        if directory is None:
            raise Exception("Directory must be provided for given URL $url")

        environment_info["directory"] = directory

        if branch is not None:
            environment_info["branch"] = branch

        environment_info["dev_paths"] = dev_paths
        environment_info["force_sync"] = force_sync
        environment_info["force_pull"] = force_pull
        environment_info["force_install"] = force_install
        environment_info["environment_hash"] = get_hash(
            url + ("" if branch is None else branch)
        )

    session_configuration["environment_info"] = environment_info

    # Upload files to S3
    for f in files:
        upload_file_to_s3(f.replace("file://", ""), s3_bucket_name)
    for f in code_files:
        upload_file_to_s3(f.replace("file://", ""), s3_bucket_name)

    session_configuration["files"] = [os.path.basename(f) for f in files]
    session_configuration["code_files"] = [os.path.basename(f) for f in code_files]

    if pf_dispatch_table is None:
        # is_it_a ? a : b in Julia becomes a if is_it_a else b in Python (edited)
        branch_to_use = (
            Repository(".").head.shorthand
            if os.getenv("BANYAN_TESTING", "0") == "1"
            else BANYAN_PYTHON_BRANCH_NAME
        )
        pf_dispatch_table = [
            "https://raw.githubusercontent.com/banyan-team/banyan-python/"
            + branch_to_use
            + "/"
            + dir
            + "/res/pf_dispatch_table.toml"
            for dir in BANYAN_PYTHON_PACKAGES
        ]

    pf_dispatch_table_loaded = load_toml(pf_dispatch_table)
    session_configuration["pf_dispatch_table"] = pf_dispatch_table_loaded
    session_configuration["language"] = "py"

    # Start the session
    response = send_request_get_response("start-session", session_configuration)
    session_id = response["session_id"]
    resource_id = response["resource_id"]

    # Store in global state
    set_session(
        session_id,
        Session(cluster_name, session_id, resource_id, nworkers, sample_rate),
    )

    wait_for_cluster(cluster_name)

    if not nowait:
        wait_for_session(session_id)

    return session_id


def end_session(
    session_id=None,
    failed=False,
    release_resources_now=False,
    release_resources_after=None,
    *args,
    **kwargs,
):
    """Ends a session given the session_id.

    Parameters
    ----------
    session_id : string
        Session ID of the session that should get ended.
        Defaults to the session_id returned from get_session_id()
    failed : bool
        Indicates whether the session being ended has failed. Defaults to False.
    release_resources_now : string
        Indicates whether to release underlying resources now. Defaults to False.
    release_resources_after: int
        The number of minutes after which to release underlying resources

    Returns
    -------
    String
        session ID of the session that was ended
    """

    configure(*args, **kwargs)

    if session_id is None:
        session_id = get_session_id()

    # Ending session with ID session_ID
    request_params = {
        "session_id": session_id,
        "failed": failed,
        "release_resources_now": release_resources_now,
    }
    if release_resources_after is not None:
        request_params["release_resources_after"] = release_resources_after
    send_request_get_response("end-session", request_params)

    # Remove from global state
    set_session(None)
    if session_id in sessions:
        del sessions[session_id]
    return session_id


def end_all_sessions(
    cluster_name,
    release_resources_now=False,
    release_resources_after=None,
    *args,
    **kwargs,
):
    """End all running sessions for a given cluster.

    Parameters
    ----------
    session_id : string
        Session ID of the session that should get ended.
        Defaults to the session_id returned from get_session_id()
    failed : bool
        Indicates whether the session being ended has failed. Defaults to False.
    release_resources_now : string
        Indicates whether to release underlying resources now. Defaults to False.
    release_resources_after: int
        The number of minutes after which to release underlying resources
    """

    configure(*args, **kwargs)

    sessions = get_sessions(cluster_name, status=["creating", "running"])
    for (session_id, session) in sessions.items():
        end_session(session_id, release_resources_now, release_resources_after)


def get_sessions(cluster_name=None, status=None, limit=-1, *args, **kwargs):
    """Gets information about all the sessions for the user. Optionally can filter
    by cluster name and status.

    Parameters
    ----------
    cluster_name : string
        Name of the cluste to filter by. Defaults to nothing
    status : string
        Status of session to filter by. Defaults to nothing

    Returns
    -------
    Dictionary
        Mappings from session ID to another dictionary containing information about the session
    """

    configure(*args, **kwargs)

    filters = {}
    if cluster_name is not None:
        filters["cluster_name"] = cluster_name

    if status is not None:
        filters["status"] = status
    # The function repeatedly calls the send_request_get_response function that takes
    # the string 'describe_sessions' and the dictionary that contains filters. looping until
    #'last_eval' does not exist in the indiv_response dictionary.
    if limit > 0:
        # Get the last `limit` number of sessions
        indiv_response = send_request_get_response(
            "describe-sessions", {"filters": filters, "limit": limit}
        )
        sessions = indiv_response["sessions"]
    else:
        # Get all sessions
        indiv_response = send_request_get_response(
            "describe-sessions", {"filters": filters}
        )
        curr_last_eval = indiv_response["last_eval"]
        sessions = indiv_response["sessions"]

        while curr_last_eval is not None:
            indiv_response = send_request_get_response(
                "describe-sessions",
                {"filters": filters, "this_start_key": curr_last_eval},
            )
            sessions.update(indiv_response["sessions"])
            curr_last_eval = indiv_response["last_eval"]

    # sessions is a dictionary that contains
    # mappings from session_id to another dictionary containing information about the session
    #                             {"start_time": "0126202220937124", "end_time": "", "num_workers": 2}
    # both start_time and end_time are of the format "yyyy-mm-dd-HH:MM:SSzzzz"
    for id, s in sessions.items():  # iterating over key and value
        if sessions[id]["end_time"] == "":
            sessions[id]["end_time"] = None
        else:
            sessions[id]["end_time"] = parse_time(sessions[id]["end_time"])
        sessions[id]["start_time"] = parse_time(sessions[id]["start_time"])
    return sessions


def get_running_sessions(*args, **kwargs):
    """Gets info about all sessions that are currently running

    Returns
    -------
    Dictionary
        Mappings from session ID to another dictionary containing information about sessions that are running
    """
    configure(*args, **kwargs)
    return get_sessions(status="running")


def get_session_status(session_id=None, *args, **kwargs):
    """Get the status of the session with the given session ID or current session
    if nothing is provided.

    Parameter
    ---------
    session_id : string
        Session ID of the session that should be got.
        Defaults to the session_id returned from get_session_id()

    Returns
    -------
    string
        Status of the session. If the status is 'failed', the 'status_explanation' is printed
    """

    configure(*args, **kwargs)

    if session_id is None:
        session_id = get_session_id()
    filters = {"session_id": session_id}  # filters is a dictionary
    # response contains a dictionary (at the key "sessions") which maps from session IDs
    # to session information where the session information is a dictionary containing
    # various info such as the status (at the key "status").
    response = send_request_get_response("describe-sessions", {"filters": filters})
    session_status = response["sessions"][session_id]["status"]
    if session_status == "failed":
        # We don't immediately fail - we're just explaining. It's only later on
        # where it's like we're actually using this session do we set the status.
        # TODO: Should this be logging.error?
        # This print statement is necessary so that we can print reason for session failure
        print(response["sessions"][session_id]["status_explanation"])
    return session_status


def download_session_logs(session_id, cluster_name, filename=None, *args, **kwargs):
    """Downloads the logs from Amazon S3 for a particular session to a local file

    Parameters
    ----------
    session_id : string
        Session ID of the session to get the log for that should get downloaded
    clusterr_name : string
        Name of the cluster the session was running on
    fileneame : string and defaults to nothing
        Path to the file on the local computer to which to download to
    """

    configure(*args, **kwargs)

    s3_bucket_name = get_cluster_s3_bucket_name(cluster_name)
    log_file_name = f"banyan-log-for-session-{session_id}"  # This creates a string with the {session_id} replaced with the value of the job_id
    if filename is None:  # if fileneame is not specified
        filename = os.path.expanduser("~")  # path to the home directory on the computer
    s3 = boto3.client("s3")
    s3.download_file(s3_bucket_name, log_file_name, filename)


def wait_for_session(session_id=None, *args, **kwargs):
    """Implements an algorithm to repeatedly get the session status and then wait for a
    period of time

    Parameters
    ----------
    session_id : string
        Session ID of the session that should get ended.
        Defaults to the session_id returned from get_session_id()

    Raises
    ------
    Raises Exception if session fails
    """

    configure(*args, **kwargs)

    if session_id is None:
        session_id = get_session_id()
    session_status = get_session_status(session_id)
    t = 5
    i = 0
    if session_status == "creating":
        pbar = tqdm(desc=f"Starting session with ID {session_id}")
    while session_status == "creating":
        time.sleep(t)
        if t < 80:
            t *= 2
        session_status = get_session_status(session_id)
        pbar.update(i)
        i += 1
    try:
        pbar.close()
    except:
        pass
    if session_status == "running":
        logging.info(f"session with ID {session_id} is ready")
    elif session_status == "completed":
        raise Exception(f"session with ID {session_id} has already completed")
    elif session_status == "failed":
        raise Exception(f"session with ID {session_id} has failed")
    else:
        raise Exception(f"Unknown session status {session_status} is ready")


def run_session(
    cluster_name=None,
    nworkers=16,
    release_resources_after=20,
    print_logs=False,
    store_logs_in_s3=True,
    store_logs_on_cluster=False,
    sample_rate=None,
    session_name=None,
    files=None,
    code_files=None,
    force_update_files=True,
    pf_dispatch_table=None,
    using_modules=None,
    url=None,
    branch=None,
    directory=None,
    dev_paths=None,
    force_sync=False,
    force_pull=False,
    force_install=False,
    estimate_available_memory=True,
    email_when_ready=None,
    *args,
    **kwargs,
):
    """Starts a session, runs some code files and the sessions ends after that."""

    if sample_rate is None:
        sample_rate = nworkers

    if files is None:
        files = []

    if code_files is None:
        code_files = []

    if using_modules is None:
        using_modules = []

    if dev_paths is None:
        dev_paths = []

    try:
        start_session(
            cluster_name=cluster_name,
            nworkers=nworkers,
            release_resources_after=release_resources_after,
            print_logs=print_logs,
            store_logs_in_s3=store_logs_in_s3,
            store_logs_on_cluster=store_logs_on_cluster,
            sample_rate=sample_rate,
            session_name=session_name,
            files=files,
            code_files=code_files,
            force_update_files=force_update_files,
            pf_dispatch_table=pf_dispatch_table,
            using_modules=using_modules,
            url=url,
            branch=branch,
            directory=directory,
            dev_paths=dev_paths,
            force_sync=force_sync,
            force_pull=force_pull,
            force_install=force_install,
            estimate_available_memory=estimate_available_memory,
            nowait=False,
            email_when_ready=email_when_ready,
            for_running=True,
        )
    except:
        try:
            session_id = get_session_id()
        except:
            session_id = None
        if session_id is not None:
            end_session(get_session_id(), failed=True)
        raise
    finally:
        try:
            session_id = get_session_id()
        except:
            session_id = None
        if session_id is not None:
            end_session(get_session_id(), failed=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['banyan']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.18.45,<2.0.0',
 'botocore>=1.23.48,<2.0.0',
 'progressbar2>=4.0.0,<5.0.0',
 'progressbar>=2.5,<3.0',
 'pygit2>=1.9.0,<2.0.0',
 'pytz>=2021.3,<2022.0',
 'requests>=2.27.1,<3.0.0',
 'toml>=0.10.2,<0.11.0',
 'tqdm>=4.63.0,<5.0.0']

setup_kwargs = {
    'name': 'banyan-python',
    'version': '0.1.0',
    'description': 'Massively parallel cloud computing with popular Python libraries for analytics, processing, and simulation! ',
    'long_description': None,
    'author': 'Banyan Computing',
    'author_email': 'support@banyancomputing.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.banyancomputing.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
